# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2013, 2020 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================
'''
Created on Jul 12, 2013
@author: Mike Stancampiano
'''
import base64
# Replacing python-crypto with pycryptodomex package
from Cryptodome.Cipher import AES
from Cryptodome import Random
from Cryptodome.Util import Counter
from Cryptodome.Hash import SHA512
import os
import random
import six
import sys
from io import BytesIO
from oslo_config.ibm import secret_opt_base_handler
from oslo_utils import encodeutils
from six.moves import configparser
BLOCK_SIZE = AES.block_size
PADDING = b'{'
ALOG_SEP = ':'


def pad(s):
    return bytes(s + (BLOCK_SIZE - len(s) % BLOCK_SIZE) * PADDING,
                 encoding='utf8')


def unpad(s):
    return s.rstrip(PADDING)

# List of valid values for encryption options.
# The first value in each list is the default value for that option


CIPHER_KEY = 'cipher'
HASH_KEY = 'hash'
PASSPHRASE_KEY = 'passphrase'
KEYSIZE_KEY = 'keysize'
MODE_KEY = 'mode'
OPTIONS_LIST = {CIPHER_KEY: ['aes'], HASH_KEY: ['sha512'],
                PASSPHRASE_KEY: [None],
                KEYSIZE_KEY: ['128', '192', '256'],
                MODE_KEY: ['ctr', 'cbc']}
# Indexes into options list
DEFAULT_INDEX = 0
AES_INDEX = 0
CTR_INDEX = 0
CBC_INDEX = 1
SHA512_INDEX = 0


class EncryptHandler(secret_opt_base_handler.SecretOptHandlerBase):
    '''
    This class provides encryption and decryption support for strings in
    PowerVC. It inherits from a base class provided by OSEE.
    '''

    _cliopts = {}

    def __init__(self, params=None):
        '''
        Parameters may be passed to an instance of this class. When an instance
        of this class is created during Oslo conf file processing no parameters
        will be passed.
        '''
        # print "Parms in Encrypthandler: %s" % params
        if params is not None:
            self._cliopts = params

    def encode(self, str_to_encrypt):
        '''
        This function takes a string as input and returns the string in
        encrypted format based on the defined encryption options.
        '''
        encrypt_opts = self.__getopts()
        # print('Encrypt options: %s' % encrypt_opts)
        if self.__chkopts(encrypt_opts):
            passphrase_hash = self._hash_key(int(encrypt_opts[KEYSIZE_KEY]),
                                             encrypt_opts[HASH_KEY],
                                             encrypt_opts[PASSPHRASE_KEY])
            if encrypt_opts[CIPHER_KEY] == OPTIONS_LIST[CIPHER_KEY][AES_INDEX]:
                return self._aes_encrypt(passphrase_hash, encrypt_opts,
                                         encodeutils.to_utf8(str_to_encrypt))
        return str_to_encrypt

    def decode(self, str_to_decrypt):
        '''
        This function takes an encrypted string as input and returns the
        decrypted string based on the defined encryption options.
        '''
        encrypt_opts = self.__getopts()
        # print('Encrypt options: %s' % encrypt_opts)
        if (self.__chkopts(encrypt_opts) and
                self._isencrypted(str_to_decrypt, encrypt_opts)):
            passphrase_hash = self._hash_key(int(encrypt_opts[KEYSIZE_KEY]),
                                             encrypt_opts[HASH_KEY],
                                             encrypt_opts[PASSPHRASE_KEY])
            msg_start = str_to_decrypt.index(ALOG_SEP) + 1
            encrypt_str = str_to_decrypt[msg_start:len(str_to_decrypt)]
            encrypt_str = base64.b64decode(encrypt_str)
            if encrypt_opts[CIPHER_KEY] == OPTIONS_LIST[CIPHER_KEY][AES_INDEX]:
                return self._aes_decrypt(passphrase_hash, encrypt_opts,
                                         encrypt_str)
        return str_to_decrypt
    
    def isencrypted(self, data):
        encrypt_opts = self.__getopts()
        return self._isencrypted(data, encrypt_opts)

    def __getopts(self):
        '''
        This function returns a dictionary of the encryption options to be
        used. It obtains the value for each encryption option based on the
        following search order:
        - Option specified during __init__
        - Option obtained from the /etc/oslo/oslo.conf file
        - Default value
        '''
        OSLO_CONF_PATH = "/etc/oslo/oslo.conf"
        encrypt_opts = self._cliopts
        config = None
        if (os.path.exists(OSLO_CONF_PATH)):
            config = configparser.RawConfigParser()
            config.read(OSLO_CONF_PATH)
        for opt_key, opt_value in six.iteritems(OPTIONS_LIST):
            if opt_key not in encrypt_opts:
                encrypt_opts[opt_key] = self.__getopt(config, opt_key,
                                                      opt_value[DEFAULT_INDEX])

        return encrypt_opts

    def __getopt(self, config, opt_str, default_value):
        '''
        This function returns the value of an encryption option found in the
        /etc/oslo/oslo.conf file or a default value
        '''
        if config:
            try:
                opt_value = config.get('DEFAULT', opt_str)
                if opt_str == PASSPHRASE_KEY:
                    try:
                        config2 = configparser.RawConfigParser()
                        config2.read(opt_value)
                        opt_value = config2.get('DEFAULT', PASSPHRASE_KEY)
                    except configparser.NoOptionError:
                        opt_value = default_value
            except configparser.NoOptionError:
                opt_value = default_value
        else:
            opt_value = default_value
        return opt_value

    def __chkopts(self, encrypt_opts):
        '''
        This function checks if the values specified for the encryption
        options are valid and returns True or False
        '''
        valid_opts = True
        for opt_key, opt_value in six.iteritems(encrypt_opts):
            valid_opt = True
            # print "Key: %s - Value: %s" % (opt_key, opt_value)
            if opt_key == PASSPHRASE_KEY and not opt_value:
                valid_opt = False
            elif (opt_key != PASSPHRASE_KEY and opt_value
                  not in (ov.lower() for ov in OPTIONS_LIST[opt_key])):
                valid_opt = False
            # print "Valid opt: %s" % valid_opt
            if not valid_opt:
                # print "Invalid value %s specifiedfor option %s" % (opt_value,
                #                                                    opt_key)
                valid_opts = False
        return valid_opts

    def _isencrypted(self, str_to_decrypt, encrypt_opts):
        '''
        This function checks if a string is encrypted based
        on the current defined encryption options and
        returns True if the string is encrypted.
        '''
        encrypt_prefix = (encrypt_opts[CIPHER_KEY] + '-' +
                          encrypt_opts[MODE_KEY] + ALOG_SEP)
        if (str_to_decrypt is None or str_to_decrypt == '' or
                str_to_decrypt[0:8] != encrypt_prefix):
            return False
        else:
            return True

    def _hash_key(self, key_size, hash_algo, passphrase):
        '''
        This function hashes a passphrase using a sha512 algorithm and
        a keysize. It returns the hashed string.
        '''
        key_hash = ''
        if hash_algo == OPTIONS_LIST[HASH_KEY][SHA512_INDEX]:
            hashobj = SHA512.new()
            hashobj.update(encodeutils.to_utf8(passphrase))
            hash_value = hashobj.hexdigest()
            key_hash = hash_value[len(hash_value) - ((key_size // 8)):
                                  len(hash_value)]
            return encodeutils.to_utf8(key_hash)

    def _aes_encrypt(self, passphrase_hash, encrypt_opts, str_to_encrypt):
        '''
        This function encrypts the passed in string using an AES cipher
        and a mode of CBC or CTR. The encrypted string is returned.
        '''
        if encrypt_opts[MODE_KEY] == OPTIONS_LIST[MODE_KEY][CBC_INDEX]:
            iv = Random.new().read(AES.block_size)
            unicode_msg = pad(str_to_encrypt)
            cipher = AES.new(passphrase_hash, AES.MODE_CBC, iv)
            encoded_str = base64.b64encode(iv + cipher.encrypt(unicode_msg))
        elif encrypt_opts[MODE_KEY] == OPTIONS_LIST[MODE_KEY][CTR_INDEX]:
            ran_gen = random.SystemRandom()
            ctr_init_value = ran_gen.randint(1, sys.maxsize)
            ctr = Counter.new(AES.block_size * 8, initial_value=ctr_init_value)
            cipher = AES.new(passphrase_hash, AES.MODE_CTR, counter=ctr)
            en_str = (str(ctr_init_value).encode('utf8') +
                      ALOG_SEP.encode() + cipher.encrypt(str_to_encrypt))
            encoded_str = base64.b64encode(en_str)
            encoded_str = encoded_str.decode('utf8')
            ecrypt_str = (encrypt_opts[CIPHER_KEY] + '-' +
                          encrypt_opts[MODE_KEY] +
                          ALOG_SEP + encoded_str)
        return ecrypt_str

    def _aes_decrypt(self, passphrase_hash, encrypt_opts, encrypt_str):
        '''
        This function decrypts the passed in string using an AES cipher
        and a mode of CBC or CTR. The decrypted string is returned.
        '''
        if encrypt_opts[MODE_KEY] == OPTIONS_LIST[MODE_KEY][CBC_INDEX]:
            iv = encrypt_str[:16]
            cipher = AES.new(passphrase_hash, AES.MODE_CBC, iv)
            unicode_msg = unpad(cipher.decrypt(encrypt_str[16:]))
        elif encrypt_opts[MODE_KEY] == OPTIONS_LIST[MODE_KEY][CTR_INDEX]:
            msg_start = encrypt_str.index(ALOG_SEP.encode()) + 1
            ctr_str = encrypt_str[0:msg_start - 1]
            ctr_init_value = int(ctr_str.decode('utf8'))
            encrypt_str = encrypt_str[msg_start:len(encrypt_str)]
            ctr = Counter.new(AES.block_size * 8, initial_value=ctr_init_value)
            cipher = AES.new(passphrase_hash, AES.MODE_CTR, counter=ctr)
            unicode_msg = cipher.decrypt(encrypt_str)
        ret_val = unicode_msg.decode('utf8')
        try:
            # attempt to convert to ascii for the sake of things that only
            # support ascii strings, else they would break
            return str(ret_val)
        except Exception as e:
            return ret_val
